/*
**	gint:thread:kernel - Thread Kernel inferface
**	This file contains all "kernel" low-level part of the thread module:
**	- thread_kernel_sched_procedure()      Thread scheduler procedure
**      - thread_kernel_terminate_trampoline() Thread terminsation trampoline
**	- thread_kernel_yield()                Relinquish the CPU
**	- thread_kernel_exit()                 Terminate the current thread
*/

.section .gint.mapped, "ax"
.align 4

.global	_thread_kernel_sched_interrupt_procedure
.type	_thread_kernel_sched_interrupt_procedure, @function

/* thread_kernel_sched_procedure(): Scheduler entry

   This "function" is involved when ANY interruption occur, this is the only way
   to add this scheduler without broke all Gint's VBR software architecture.

   The scheduler use a ~128hz timer to schedule thread contexts. During the
   "driver" initialization (yes, the thread module is considered like a driver
   for Gint's eyes) one timer has been locked and all hardware information about
   it has been pre-calculated during this part.

   @note
   Due to the current Gint's VBR architecture, this function is involved each
   time an interruption occur. By the fact that this function is involved like
   a common subroutine, PR register has been saved into the R0 register and the
   R15 register has been saved into the R1 register.

   We are always with the "privileged" bank register configuration. So, we only
   can use r0~r7 register until the current thread context has been saved.

   When involved, R0 content the PR register snapshot when interrupts occur,
   same for the R1 register which content the stack snapshot. (Yes, I know that
   the SH4-based MPU provide the SGR register which saves the stack snapshot
   when interrupts / exceptions occur, but we need to be SH3 compatible) */
_thread_kernel_sched_interrupt_procedure:
	/* First, check if the interruption is the scheduler's timer */
	mov.l	intevt_register, r2
	mov.l	thread_sched_tmu_interrupt_id, r3
	mov.l	@r2, r2
	mov.l	@r3, r3
	cmp/eq	r2, r3
	bt	_thread_kernel_sched_entry
	rts
	nop

_thread_kernel_sched_entry:
	/* Stop the scheduler's timer to "avoid" counting the context
	   switching time. */
	/* The scheduler can use ETMU instead of a classical TMU. ETMU have
	   slow writting interraction, but have the same access size that the
	   TMU, this is why we use slow operation here without checking the
	   timer type. */
	mov.l	thread_sched_tmu_tstr_addr, r2
	mov.l	thread_sched_tmu_tstr_mask, r3
	mov.l	@r2, r2
	mov.l	@r3, r3
	mov.b	@r2, r6
	not	r3, r3
	and	r3, r6
timer_tstr_slow_stop:
	mov.b	r6, @r2
	mov.b	@r2, r7
	cmp/eq	r6, r7
	bf	timer_tstr_slow_stop

save_context:
	/* check if a current thread is running */
	mov.l	thread_sched_current, r2
	mov.l	@r2, r2
	tst	r2, r2
	bt/s	schedule_thread

	/* save current thread context */
	add	#88, r2
	mov.l	r0, @-r2	/* R0 contains user PR register snapshot */
	stc.l	spc, @-r2
	stc.l	ssr, @-r2
	sts.l	mach, @-r2
	sts.l	macl, @-r2
	stc.l	gbr, @-r2
	mov.l	r1, @-r2	/* R1 contains user R15 register snapshot */
	mov.l	r14, @-r2
	mov.l	r13, @-r2
	mov.l	r12, @-r2
	mov.l	r11, @-r2
	mov.l	r10, @-r2
	mov.l	r9, @-r2
	mov.l	r8, @-r2
	stc.l	R7_BANK, @-r2
	stc.l	R6_BANK, @-r2
	stc.l	R5_BANK, @-r2
	stc.l	R4_BANK, @-r2
	stc.l	R3_BANK, @-r2
	stc.l	R2_BANK, @-r2
	stc.l	R1_BANK, @-r2
	stc.l	R0_BANK, @-r2

schedule_thread:
	/* Call high-level abraction

	   We need to save the R1/R0 registers (that contains the user stack/PR
	   register snapshot respectively) because if the scheduler fail, we
	   SHOULD be able to restore the context before returning from here.

	   We alse need to swtich the stack to avoid undefined behaviour if the
	   current thread is destroyed during the schedule (we currently always
	   use the thread stack here) */
	mov.l	thread_kernel_stack, r15
	mov.l	r1, @-r15
	mov.l	r0, @-r15
	mov.l	thread_schedule, r0
	jsr	@r0
	nop
#ifdef THREAD_SCHEDULER_DEBUG
	mov.l	thread_schedule_debug, r1
	jsr	@r1
	mov	r0, r4
#endif
	tst	r0, r0
	lds.l	@r15+, pr
	bt/s	scheduler_restart_timer
	mov.l	@r15+, r15


context_restore:
	/* update_current_thread */
	mov.l	thread_sched_current, r1
	mov.l	r0, @r1

	/* restore the new context */
	ldc.l	@r0+, R0_BANK
	ldc.l	@r0+, R1_BANK
	ldc.l	@r0+, R2_BANK
	ldc.l	@r0+, R3_BANK
	ldc.l	@r0+, R4_BANK
	ldc.l	@r0+, R5_BANK
	ldc.l	@r0+, R6_BANK
	ldc.l	@r0+, R7_BANK
	mov.l	@r0+, r8
	mov.l	@r0+, r9
	mov.l	@r0+, r10
	mov.l	@r0+, r11
	mov.l	@r0+, r12
	mov.l	@r0+, r13
	mov.l	@r0+, r14
	mov.l	@r0+, r15
	ldc.l	@r0+, gbr
	lds.l	@r0+, macl
	lds.l	@r0+, mach
	ldc.l	@r0+, ssr
	ldc.l	@r0+, spc
	lds.l	@r0+, pr

scheduler_restart_timer:
	/* Check if the scheduler uses TMU or ETMU */
	mov.l	thread_sched_gint_timer_id, r0
	mov.l	@r0, r0
	mov	#3, r1
	cmp/ge	r1, r0

	/* Get scheduler timer pre-calculated information */
	mov.l	thread_sched_tmu_tcr_addr, r0
	mov.l	thread_sched_tmu_tcr_mask, r1
	mov.l	@r0, r0
	mov.l	@r1, r1
	not	r1, r1
	mov.l	thread_sched_tmu_tcor_addr, r2
	mov.l	thread_sched_tmu_tcnt_addr, r3
	mov.l	@r2, r2
	mov.l	@r3, r3
	mov.l	thread_sched_tmu_tstr_addr, r4
	mov.l	thread_sched_tmu_tstr_mask, r5
	mov.l	@r4, r4
	bf/s	tmu_restart_timer
	mov.l	@r5, r5

etmu_restart_timer:
	/* clear interrupt flag (ETMU.TCR.UNF = 0) */
	mov.b	@r0, r6
	and	r1, r6
etmu_tcr_slow_unf_clear:
	mov.b	r6, @r0
	mov.b	@r0, r7
	cmp/eq	r6, r7
	bf	etmu_tcr_slow_unf_clear

	/* reload the timer counter (ETMU.TCNT = ETMU.TCOR) */
	mov.l	@r2, r6
etmu_slow_reload:
	mov.l	r6, @r3
	mov.l	@r3, r7
	cmp/eq	r6, r7
	bf	etmu_slow_reload

	/* Check if we really need to restart the timer or not */
	mov.l	thread_sched_counter, r0
	mov.l	@r0, r0
	mov	#2, r1
	cmp/ge	r1, r0
	bf	process_switch

	/* Restart timer (ETMU.TSTR = 1) */
	mov.b	@r4, r6
	or	r5, r6
etmu_tstr_slow_restart:
	mov.b	r6, @r4
	mov.b	@r4, r7
	cmp/eq	r6, r7
	bf	etmu_tstr_slow_restart
	bra	process_switch
	nop

tmu_restart_timer:
	/* clear interrupt flag (TMU.TCR.UNF) */
	mov.w	@r0, r6
	and	r1, r6
	mov.w	r6, @r0

	/* reload the timer counter */
	mov.l	@r2, r6
	mov.l	r6, @r3

	/* Check if we really need to restart the timer or not */
	mov.l	thread_sched_counter, r0
	mov.l	@r0, r0
	mov	#2, r1
	cmp/ge	r1, r0
	bf	process_switch

	/* Restart timer (TMU.TSTR) */
	mov.b	@r4, r6
	or	r5, r6
	mov.b	r6, @r4

process_switch:
	rte
	nop

.align 4
/* kernel information */
thread_kernel_stack:		.long _thread_kernel_stack_start

/* scheduler global information */
thread_sched_current:		.long _thread_sched_current
thread_sched_counter:		.long _thread_sched_counter

/* High-level functions */
thread_schedule:		.long _thread_schedule
#ifdef THREAD_SCHEDULER_DEBUG
thread_schedule_debug:		.long _thread_schedule_debug
#endif

/* Timer hardware pre-calculated information */
thread_sched_gint_timer_id:	.long _thread_sched_gint_timer_id
thread_sched_tmu_interrupt_id:	.long _thread_sched_tmu_interrupt_id
thread_sched_tmu_tcr_addr:	.long _thread_sched_tmu_tcr_addr
thread_sched_tmu_tcr_mask:	.long _thread_sched_tmu_tcr_mask
thread_sched_tmu_tstr_mask:	.long _thread_sched_tmu_tstr_mask
thread_sched_tmu_tstr_addr:	.long _thread_sched_tmu_tstr_addr
thread_sched_tmu_tcor_addr:	.long _thread_sched_tmu_tcor_addr
thread_sched_tmu_tcnt_addr:	.long _thread_sched_tmu_tcnt_addr

/* Other information */
intevt_register:		.long 0xff000028




.global	_thread_kernel_yield
.type	_thread_kernel_yield, @function

/* thread_kernel_yield(): Cause the calling thread to relinquish the CPU

   This function will move the calling thread to the end of the queue for its
   static priority and new thread gets on run. */
/* TODO preemption ? */
_thread_kernel_yield:
	/* start atomic operation + bank switch*/
	stc	sr, r0
	mov	r0, r1
	mov	#0x30, r2
	shll8	r2
	shll16	r2
	or	r2, r1
	ldc 	r1, sr

	/* prepare bank switch */
	stc	R0_BANK, r0
	sts	pr, r1
	ldc	r0, ssr
	ldc	r1, spc

	/* simulate the interrupt by switching the register bank */
	mov	#0x20, r1
	shll8	r1
	shll16	r1
	or	r1, r0
	ldc	r0, sr

	/* prepare to jump into thread_kernel_scheudler_entry() */
	xor	r0, r0
	mov	r15, r1
	mov.l	1f, r2
	jsr	@r2
	nop

pouet:
	sleep
	bra	pouet
	nop



.global	_thread_kernel_terminate_trampoline
.type	_thread_kernel_terminate_trampoline, @function

/* thread_kernel_terminate_trampoline()
    call the thread_exit function with the returned value.*/
_thread_kernel_terminate_trampoline:
	mov.l	2f, r1
	jmp	@r1
	mov	r0, r4

.align 4
1:	.long _thread_kernel_sched_entry
2:	.long _thread_exit
