//---
//	gint:thread:mutex - Thread mutex API
//---
#include <gint/thread.h>
#include <gint/timer.h>
#include <gint/std/string.h>

//---
//	Internal functions
//---
/* thread_mutex_timer_callback()

   Involved when a "thread_timedlock()" timeout is reached. This function will
   request to abort the locking try. */
static int thread_mutex_timer_callback(thread_mutex_t *mutex)
{
	thread_atomic_start();
	mutex->timer.abord = 1;
	thread_atomic_stop();
	return (TIMER_STOP);
}

/* thread_mutex_check_validity(): Check the validity of the mutex

   This function will check the watermark and the thread owner. It returns the
   associated thread owner on success, NULL otherwise.

   This function SHOUL be involved in "atomic" mode. */
static struct thread *thread_mutex_validity_check(thread_mutex_t *mutex)
{
	struct thread *thread;

	if (mutex == NULL || mutex->watermark != THREAD_MUTEX_WATERMARK)
		return (NULL);
	thread = thread_sched_get_current();
	if (thread == NULL)
		return (NULL);
	return (thread);
}

//---
//	Thread mutex API
//---

/* thread_mutex_init(): Creates a new mutex object with type TYPE */
int thread_mutex_init(thread_mutex_t *mutex, int type)
{
	thread_atomic_start();
	if (mutex->watermark == THREAD_MUTEX_WATERMARK) {
		thread_atomic_stop();
		return (thread_mutex_retval_success);
	}
	switch (type) {
	case thread_mutex_type_plain:
	case thread_mutex_type_timed:
	case thread_mutex_type_timed | thread_mutex_type_recursive:
	case thread_mutex_type_plain | thread_mutex_type_recursive:
		break;
	default:
		thread_atomic_stop();
		return (thread_mutex_retval_error);
	}
	memset(mutex, 0x00, sizeof(thread_mutex_t));
	mutex->watermark = THREAD_MUTEX_WATERMARK;
	mutex->type = type;
	mutex->lock = 0;
	mutex->timer.id = -1;
	mutex->timer.abord = 0;
	mutex->owner = -1;
	thread_atomic_stop();
	return (thread_mutex_retval_success);
}

/* thread_mutex_lock(): Block the current thread until the mutex is locked */
int thread_mutex_lock(thread_mutex_t *mutex)
{
	struct thread *thread;

	/* Check error */
	thread_atomic_start();
	thread = thread_mutex_validity_check(mutex);
	if (thread == NULL) {
		thread_atomic_stop();
		return (thread_mutex_retval_error);
	}

	/* Check if the mutex is recursive */
	if ((mutex->type & thread_mutex_type_recursive) != 0) {
		if (mutex->owner == thread->scheduler.id) {
			mutex->lock = mutex->lock + 1;
			thread_atomic_stop();
			return (thread_mutex_retval_success);
		}
	}

	/* Wait util the mutex is locked */
	while (1) {
		/* Check if the mutex is unlocked */
		if (mutex->lock == 0) {
			mutex->owner = thread->scheduler.id;
			mutex->lock = 1;
			break;
		};

		/* Force schedule */
		thread_kernel_yield();
	}

	/* Lock and return */
	thread_atomic_stop();
	return (thread_mutex_retval_success);
}

int thread_mutex_timedlock(thread_mutex_t *mutex, uint64_t delay_us)
{
	struct thread *thread;
	int retval;

	/* Check error */
	thread_atomic_start();
	thread = thread_mutex_validity_check(mutex);
	if (thread == NULL || (mutex->type & thread_mutex_type_timed) == 0) {
		thread_atomic_stop();
		return (thread_mutex_retval_error);
	}

	/* Check if the mutex is recursive */
	if (mutex->type & thread_mutex_type_recursive) {
		if (mutex->owner == thread->scheduler.id) {
			mutex->lock = mutex->lock + 1;
			thread_atomic_stop();
			return (thread_mutex_retval_success);
		}
	}

	/* setup the timer */
	mutex->timer.abord = 0;
	mutex->timer.id = timer_setup(TIMER_ANY, delay_us,
				(void*)&thread_mutex_timer_callback, mutex);
	if (mutex->timer.id < 0) {
		thread_atomic_stop();
		return (thread_mutex_retval_error);
	}

	/* Wait util the mutex is unlocked or timeout */
	timer_start(mutex->timer.id);
	while (1) {
		/* Check timeout abord */
		if (mutex->timer.abord == 1) {
			retval = thread_mutex_retval_busy;
			break;
		}

		/* Check if the mutex is unlocked */
		if (mutex->lock == 0) {
			mutex->lock = 1;
			mutex->owner = thread->scheduler.id;
			retval = thread_mutex_retval_success;
			break;
		};

		/* Force schedule */
		thread_kernel_yield();
	}

	/* destroy the timer and return */
	timer_stop(mutex->timer.id);
	mutex->timer.id = -1;
	mutex->timer.abord = -1;
	thread_atomic_stop();
	return (retval);
}

/* thread_mutext_trylock(): Try to lock the mutex without blocking. */
int thread_mutex_trylock(thread_mutex_t *mutex)
{
	struct thread *thread;

	/* Check error */
	thread_atomic_start();
	thread = thread_mutex_validity_check(mutex);
	if (thread == NULL) {
		thread_atomic_stop();
		return (thread_mutex_retval_error);
	}

	/* Check recursive lock */
	if (mutex->type & thread_mutex_type_recursive) {
		if (mutex->owner == thread->scheduler.id) {
			mutex->lock = mutex->lock + 1;
			thread_atomic_stop();
			return (thread_mutex_retval_success);
		}
	}

	/* Check if the mutex is unlocked */
	if (mutex->lock == 0) {
		mutex->owner = thread->scheduler.id;
		mutex->lock = 1;
		thread_atomic_stop();
		return (thread_mutex_retval_success);
	};

	/* the mutex is busy */
	thread_atomic_stop();
	return (thread_mutex_retval_busy);
}

/* thread_mutex_unlock(): Try to unlock the mutex */
int thread_mutex_unlock(thread_mutex_t *mutex)
{
	struct thread *thread;

	thread_atomic_start();
	thread = thread_mutex_validity_check(mutex);
	if (thread == NULL
	|| mutex->owner != thread->scheduler.id
	|| mutex->lock == 0) {
		thread_atomic_stop();
		return (thread_mutex_retval_error);
	}
	mutex->lock = mutex->lock - 1;
	thread_atomic_stop();
	return (thread_mutex_retval_success);
}

/* thread_mutex_destroy(): Destroy the mutex object */
void thread_mutex_destroy(thread_mutex_t *mutex)
{
	thread_atomic_start();
	mutex->watermark = 0x00000000;
	mutex->type  = 0;
	mutex->owner = -2;
	mutex->lock  = 0;
	if (mutex->timer.id >= 0)
		timer_stop(mutex->timer.id);
	thread_atomic_stop();
}
