//---
//	gint:thread:scheduler - Scheduler module
//---
#include <gint/thread.h>
#include <gint/timer.h>
#include <gint/std/stdlib.h>

#include <gint/display.h>


/* define symbols that will be used bu the kernel to communicate with us */
struct thread *thread_sched_queue;
struct thread *thread_sched_current;
uint32_t thread_sched_counter;
thread_t thread_sched_uuid;

/* define symbols used to pre-calculate scheudler timer related information */
uint32_t thread_sched_tmu_interrupt_id;
uint32_t thread_sched_tmu_tcr_addr;
uint32_t thread_sched_tmu_tcr_mask;
uint32_t thread_sched_tmu_tstr_mask;
uint32_t thread_sched_tmu_tstr_addr;
uint32_t thread_sched_tmu_tcor_addr;
uint32_t thread_sched_tmu_tcnt_addr;
int thread_sched_gint_timer_id;


//---
//	Driver primitives
//---

/* thread_sched_init(): Initialize the scheduler */
void thread_sched_init(void)
{
	struct timer_debug_info info;
	uint64_t  delay;

	/* initialize kernel information */
	thread_sched_queue = NULL;
	thread_sched_current = NULL;
	thread_sched_counter = 0;
	thread_sched_uuid = 0;

	/* generate the delay in us */
	delay = 1000000 / THREAD_SCHEDULER_FREQUENCY;
	if(delay == 0)
		delay = 1;

	/* try to lock one timer */
	thread_sched_gint_timer_id = timer_setup(TIMER_ANY, delay, NULL);
	if (thread_sched_gint_timer_id < 0)
		return;

	/* pre-calculate timer information, used by the
	   "thread_kernel_sched_enty()" kernel function (see
	   <gint/thread/kernel.S> for more information ) */
	timer_debug_get_hw_info(thread_sched_gint_timer_id, &info);
	thread_sched_tmu_tcr_addr  = (uintptr_t)info.hardware.address.tcr;
	thread_sched_tmu_tcr_mask  = (uintptr_t)info.interrupt.mask.tcr.unf;
	thread_sched_tmu_tstr_addr = (uintptr_t)info.hardware.address.tstr;
	thread_sched_tmu_tstr_mask = (uintptr_t)info.interrupt.mask.tstr.str;
	thread_sched_tmu_tcor_addr = (uintptr_t)info.hardware.address.tcor;
	thread_sched_tmu_tcnt_addr = (uintptr_t)info.hardware.address.tcnt;
	thread_sched_tmu_interrupt_id = (uintptr_t)info.interrupt.id;
}

/* thread_shced_uninit(): Uninitialize the scheduler */
void thread_sched_uninit(void)
{
	struct thread **thread;
	struct thread *tmp;

	thread_atomic_start();
	thread_sched_stop();
	thread = &thread_sched_queue;
	while (*thread != NULL) {
		tmp = *thread;
		*thread = (*thread)->scheduler.next;
		//thread_kill(tmp, 0);
		thread_sched_remove(tmp);
	}
}




//---
//	User interface
//---
/* thread_sched_start(): Start the scheduler timer */
void thread_sched_start(void)
{
	thread_atomic_start();
	timer_start(thread_sched_gint_timer_id);
	thread_atomic_stop();
}

/* thread_shced_stop(): */
void thread_sched_stop(void)
{
	thread_atomic_start();
	timer_pause(thread_sched_gint_timer_id);
	thread_atomic_stop();
}

/* thread_sched_add(): Add thread to the scheduler queue */
int thread_sched_add(struct thread *thread)
{
	thread_atomic_start();
	/* link the thread */
	thread->scheduler.id = thread_sched_uuid;
	thread->scheduler.status = THREAD_STATUS_RUNNING;
	thread->scheduler.next = thread_sched_queue;
	thread_sched_queue = thread;

	/* update internal information */
	thread_sched_uuid = thread_sched_uuid + 1;
	thread_sched_counter = thread_sched_counter + 1;
	thread_atomic_stop();
	return (0);
}

/* thread_sched_remove(): Add thread to the scheduler queue */
int thread_sched_remove(struct thread *thread)
{
	struct thread **parent;

	/* Try to find the thread's parent */
	thread_atomic_start();
	parent = &thread_sched_queue;
	while (*parent != NULL) {
		if (*parent == thread)
			break;
		parent = &(*parent)->scheduler.next;
	}
	if (*parent == NULL) {
		thread_atomic_stop();
		return (-1);
	}

	/* unlink the thread */
	*parent = thread->scheduler.next;

	/* remove scheduler information */
	thread->scheduler.id = -1;
	thread->scheduler.next = NULL;

	/* update internal information */
	thread_sched_counter = thread_sched_counter - 1;
	thread_atomic_stop();
	return (0);
}

/* thread_sched_get_current(): Get the current thread */
struct thread *thread_sched_get_current(void)
{
	return(thread_sched_current);
}


//---
//	Kernel interface
//---
/* thread_sched_check(): Check thread validity

   @return:
   * 0   can be used
   * -1  cannot be loaded
   * -2  has been removed
*/
static int thread_sched_check(struct thread *thread)
{
	if (thread->scheduler.status != THREAD_STATUS_RUNNING)
			return (-1);
	switch (thread_signals_pending_deliver(thread)) {
	case thread_signals_deliver_retval_running:	return (0);
	case thread_signals_deliver_retval_stopped:	return (-1);
	case thread_signals_deliver_retval_dead:
	default:
		if (thread_terminate(thread) == 0)
			return (-2);
	}
	return (-1);
}

/* thread_schedule(): Schedule thread queue and return the next thread

   This function SHOULD NOT be involved manually, it should be be involved only
   by the scheduler "handler" (see <gint/thread/kernel.S>). If you know what
   you are doing and whant to call this function, you SHOULD do it in "atomic"
   environment. */
struct thread *thread_schedule(void)
{
	struct thread *thread;
	struct thread *next;

	/* Check potential error. */
	if (thread_sched_queue == NULL)
		return (NULL);

	/* if we have no currently running thread, return directly */
	if (thread_sched_current == NULL)
		return (thread_sched_queue);

	/* pre-calculate the next thread */
	next = thread_sched_current;
	if (thread_sched_current == NULL)
		next = thread_sched_queue;
	next = next->scheduler.next;
	if (next == NULL)
		next = thread_sched_queue;

	do {
		/* Get the potential next thread because the current can be
		   destroyed during operations */
		thread = next;
		next = next->scheduler.next;
		if (next == NULL)
			next = thread_sched_queue;

		/* Check the thread validity */
		if (thread_sched_check(thread) == 0)
			return (thread);
	} while (thread != thread_sched_current);

	/* If no thread has been found, load idle kernel thread which
	will only wait the next scheduler timer intervention.
	(see <gint/thread/thread.c> for more information). */
	return (thread_idle_get());
}

/* thread_sched_find(): Find the thread using its identifier */
struct thread *thread_sched_find(thread_t id)
{
	struct thread *thread;

	thread_atomic_start();
	thread = thread_sched_queue;
	while (thread != NULL) {
		if (thread->scheduler.id == id) {
			thread_atomic_stop();
			return (thread);
		}
		thread = thread->scheduler.next;
	}
	thread_atomic_stop();
	return (NULL);
}

/* invalidate the current thread to avoid context saving */
/* TODO: find better way to do the job */
/* TODO: doc */
void thread_sched_invalidate(void)
{
	thread_sched_current = NULL;
}




//---
//	Debugging interface
//---
#ifdef THREAD_SCHEDULER_DEBUG
#include <gint/display.h>

int context_switch_counter = 0;

/* thread_schedule_debug(): Debug scheduler, involved each context_switch */
struct thread *thread_schedule_debug(struct thread *thread)
{
/*	extern uint32_t thread_tmu_interrupt_id;
	extern uint32_t thread_tmu_tcr_addr;
	extern uint32_t thread_tmu_tcr_mask;
	extern uint32_t thread_tmu_tstr_mask;
	extern uint32_t thread_tmu_tstr_addr;

	dclear(C_WHITE);
	dprint(1, 51,  C_BLACK, "next process -> %p", thread);
 	dprint(1, 61, C_BLACK, "|--spc: %p", thread->context.cpu.spc);
	dprint(1, 71, C_BLACK, "|--ssr: %p", thread->context.cpu.ssr);
	dprint(1, 81, C_BLACK, "|--r15: %p", thread->context.cpu.reg[15]);
	dprint(1, 91, C_BLACK, "|--@r0: %p", &thread->context.cpu.reg[0]);
	dprint(1, 101, C_BLACK, "|--tmu_event: %#x", thread_tmu_interrupt_id);
	dprint(1, 111, C_BLACK, "|--tmu_tcr_addr: %#x", thread_tmu_tcr_addr);
	dprint(1, 121, C_BLACK, "|--tmu_tcr_mask: %#x", thread_tmu_tcr_mask);
	dprint(1, 131, C_BLACK, "|--tmu_tstr_addr: %#x", thread_tmu_tstr_addr);
	dprint(1, 141, C_BLACK, "`--tmu_tstr_mask: %#x", thread_tmu_tstr_mask);
*/

/*
	struct thread **tmp;
	int idx;
	dclear(C_WHITE);
	dprint(1, 51, C_BLACK, "new process -> %p", thread);
	dtext (1, 61, C_BLACK, "thread queue:");
	idx = 0;
	tmp = &thread_scheduler_queue;
	while (*tmp != NULL) {
		dprint(1, 71 + (10 * idx), C_BLACK, "[%d] - %p", idx, *tmp);
		tmp = &(*tmp)->private.next;
		idx = idx + 1;
	}
	dupdate();
	if (idx != 1)
		while (1);
*/

	context_switch_counter = context_switch_counter + 1;
//	if (context_switch_counter < 3)
		return (thread);

	uint8_t tstr_before = ((uint8_t *)thread_sched_tmu_tstr_addr)[0];
	uint8_t tstr_after  = ((uint8_t *)thread_sched_tmu_tstr_addr)[0] & (~thread_sched_tmu_tstr_mask);

	dclear(C_WHITE);
	dprint(1, 11,  C_BLACK, "next process -> %p", thread);
 	dprint(1, 21, C_BLACK, "TMU -> ID:   %#x", (uint8_t *)thread_sched_gint_timer_id);
 	dprint(1, 31, C_BLACK, "TMU -> TSTR: %#x", tstr_before);
	dprint(1, 41, C_BLACK, "TMU -> TCOR: %#x", ((uint32_t*)thread_sched_tmu_tcor_addr)[0]);
	dprint(1, 51, C_BLACK, "TMU -> TCNT: %#x", ((uint32_t*)thread_sched_tmu_tcnt_addr)[0]);
	dprint(1, 71, C_BLACK, "TMU -> TSTR %#x", thread_sched_tmu_tstr_mask);
	dprint(1, 81, C_BLACK, "TMU -> TSTR %#x", tstr_after);
	dupdate();
	while (1);
}
#endif
