//---
//	gint:thread:signal - Thread Signals management
//---
#include <gint/thread.h>
#include <gint/std/string.h>

//---
//	internals functions
//---
/* thread_signals_deliver(): Deliver a signal */
static int thread_signals_deliver(struct thread *thread, int sig)
{
	// check if the signal is ignored
	if (thread->signals.handler[sig] == SIG_IGN)
		return (thread_signals_deliver_retval_running);

	// Check custom signal managment
	if (thread->signals.handler[sig] != SIG_DFL) {
		/* save current context into user's stack and update stack */
		thread->context.reg[15] -= sizeof(struct cpu_ctx) + 3;
		thread->context.reg[15]  = thread->context.reg[15] >> 2 << 2;
		memcpy((void*)(uintptr_t)thread->context.reg[15],
				&thread->context, sizeof(struct cpu_ctx));

		/* clean context and set the handler */
		for (int i = 0 ; i < 14 ; i = i + 1)
			thread->context.reg[i] = 0x00000000;
		thread->context.gbr  = 0x00000000;
		thread->context.macl = 0x00000000;
		thread->context.mach = 0x00000000;
		thread->context.ssr  = 0x40000000;
		thread->context.spc  = (uintptr_t)thread->signals.handler[sig];
		thread->context.pr   = (uintptr_t)&thread_signals_sigreturn;

		/* send the signal number */
		thread->context.reg[4] = sig;
		return (thread_signals_deliver_retval_running);
	}

	/* default actions */
	switch (sig) {
			/* Stop the current thread (TODO: stop all child ?) */
		case SIGSTOP:
			thread->scheduler.status = THREAD_STATUS_PAUSED;
			return (thread_signals_deliver_retval_stopped);

			/* Wake up the parent */
		case SIGCONT:
			if (thread->private.parent != NULL) {
				thread_signals_raise(
					thread->private.parent, SIGCHLD);
			}
			return (0);

			/* Nothing, just wake up the process */
		case SIGCHLD:
			return (0);

			/* Process terination (default) */
		case SIGTERM:	// termination (TODO: cleanup part !!!)
		case SIGINT:	// interruption
		case SIGHUP:	// hang up
		case SIGKILL:	// kill
		case SIGBUS:	// bus error
		case SIGFPE:	// fatal arithmetic error
		case SIGSEGV:	// segmentation violation
		default:
			return (thread_signals_deliver_retval_dead);
	}
}

//---
//	Kernel interface
//---
/* thread_signals_raise():Raise a signal */
int thread_signals_raise(struct thread *thread, int sig)
{
	if (sig >= NSIG)
		return (-1);
	thread_atomic_start();
	if (thread == NULL
	||  thread->scheduler.status == THREAD_STATUS_ZOMBIE) {
		thread_atomic_stop();
		return (-1);
	}
	thread->scheduler.status = THREAD_STATUS_RUNNING;
	thread->signals.pending |= 1 << sig;
	thread_atomic_stop();
	return (0);
}

/* thread_signals_deliver_pending(): Deliver pending signals */
int thread_signals_pending_deliver(struct thread *thread)
{
	sigset_t sig;
	sigset_t block;
	int retval;

	retval = 0;
	thread_atomic_start();
	block = thread->signals.blocking;
	block = block & ~(1 << SIGKILL);
	block = block & ~(1 << SIGSTOP);
	block = block & ~(1 << SIGTERM);
	sig = thread->signals.pending & (~block);
	for (int i = 0; i < NSIG && retval == 0; ++i) {
		if ((sig & (1 << i)) == 0)
			continue;
		retval = thread_signals_deliver(thread, i);
		thread->signals.pending &= ~(1 << i);
	}
	thread_atomic_stop();
	return (retval);
}

/* thread_signals_sigreturn(): Signals return handler */
void thread_signals_sigreturn(void)
{
	struct thread *thread;
	void *stack;

	/* get the current thread */
	thread_atomic_start();
	thread = thread_sched_get_current();
	if (thread == NULL) {
		thread_atomic_stop();
		while (1) { thread_kernel_yield(); };
	}

	/* dump old context */
	stack = (void*)(uintptr_t)thread->context.reg[15];
	memcpy(&thread->context, stack, sizeof(struct cpu_ctx));

	/* restore stack */
	thread->context.reg[15] += (sizeof(struct cpu_ctx) + 3) >> 2 << 2;

	/* Invalidate the thread to force the scheduler to not save the current
	   context otherwise, the restored context will be overwritten. */
	thread_sched_invalidate();

	/* force schedule, on the next schedule, we will return to the job */
	thread_atomic_stop();
	while (1) {
		thread_kernel_yield();
	}
}

/* thread_signals_replace(): Replace current signal handler */
void (*thread_signals_replace(struct thread *thread,
					int signum, void (*handler)(int)))(int)
{
	void (*old)(int);

	if (signum >= NSIG)
		return (SIG_ERR);
	thread_atomic_start();
	if (thread == NULL
	||  thread->scheduler.status == THREAD_STATUS_ZOMBIE) {
		thread_atomic_stop();
		return (SIG_ERR);
	}
	old = thread->signals.handler[signum];
	thread->signals.handler[signum] = handler;
	thread_atomic_stop();
	return (old);
}
