//---
//	gint:thread:thread - Thread management
//---
#include <gint/thread.h>
#include <gint/timer.h>
#include <gint/drivers.h>
#include <gint/std/setjmp.h>
#include <gint/std/stdlib.h>
#include <gint/std/string.h>

#include <gint/display.h>

#undef thread_create

//---
//	Internal thread actions
//---

#if 0
/* thread_kill_child()

   If the THREAD_ATTR_MAIN_THREAD is set, this function is involved when the
   thread die. This function will kill all child thread generated by the parent
   */
/* TODO */
static void thread_kill_child(struct thread **thread)
{
	/* walking */
	if (*thread == NULL)
		return;
	thread_kill_child(&(*thread)->private.child);
	thread_kill_child(&(*thread)->private.sibling);

	/* destroy the thread */
	/*thread_atomic_start();
	thread_kill(*thread, 0);
	thread_sched_remove(*thread);
	*thread = NULL;*/
	thread_atomic_stop();
}
#endif



//---
//	User thread interface
//---

/* thread_create(): Create a new thread */
int thread_create(thread_t *tid, thread_attr_t *attr, void *function, ...)
{
	struct thread *thread;
	struct thread *parent;
	uint32_t argbuf[16];
	int arg_counter;
	uintptr_t arg;
	va_list ap;

	/* Start the scheduler to avoid small freeze with the first thread */
	thread_atomic_start();
	thread_sched_start();

	/* create the new thread */
	thread = (struct thread *)calloc(1, sizeof(struct thread));
	if (thread == NULL) {
		thread_atomic_stop();
		return (-1);
	}
	thread->private.stack = (uintptr_t)malloc(THREAD_STACK_SIZE + 4);
	if (thread->private.stack == 0x00000000) {
		free(thread);
		thread_atomic_stop();
		return (-1);
	}
	thread->context.reg[15]  = ((thread->private.stack + 3) >> 2 << 2);
	thread->context.reg[15] += ((THREAD_STACK_SIZE + 3) >> 2 << 2);
	thread->context.pr  = (uintptr_t)thread_kernel_terminate_trampoline;
	thread->context.spc = (uintptr_t)function;
	thread->context.ssr = 0x40000000;

	/* dump arguments */
	va_start(ap, function);
	for (arg_counter = 0; arg_counter <= 16; ++arg_counter) {
		arg = va_arg(ap, uint32_t);
		if (arg == THREAD_CREATE_ARG_END_WATERMARK)
			break;
		if (arg_counter < 4) {
			thread->context.reg[arg_counter + 4] = arg;
		} else {
			argbuf[arg_counter - 4] = arg;
		}
	}
	va_end(ap);
	if (arg_counter >= 4) {
		arg_counter -= 4;
		thread->context.reg[15] -= arg_counter << 2;
		memcpy((void*)(uintptr_t)thread->context.reg[15],
						argbuf, arg_counter << 2);
	}

	/* dump attribute if needed */
	if (attr != NULL) {
		memcpy(&thread->attr, attr, sizeof(struct thread_attr_s));

		/* check longjmp() feature */
		if (attr->state.longjmp == THREAD_ATTR_LONGJMP_ENABLE) {
			if (setjmp(thread->private.jmpbuf) != 0) {
				thread_atomic_stop();
				return (THREAD_CREATE_LONGJMP_RETURN);
			}
		}

		/* check joinable state */
		if (attr->state.detach == THREAD_ATTR_JOINABLE) {
			parent = thread_sched_get_current();
			if (parent != NULL) {
				thread->private.parent = parent;
				thread->private.sibling = parent->private.child;
				parent->private.child = thread;
			}
		}
	}

	/* link the thread to the scheduler */
	thread_sched_add(thread);
	*tid = thread->scheduler.id;

	/* initialize signals */
	for (int i = 0; i < NSIG; ++i)
		thread->signals.handler[i] = SIG_DFL;
	thread->signals.pending = 0x00000000;
	thread->signals.blocking = 0x00000000;
	thread_atomic_stop();
	return (0);
}

/* thread_tryjoin()
   Same behaviour than thread_join() except that if the thread is not in the
   ZOMBIE_STATE its return directly with an error. */
int thread_tryjoin(thread_t id, void **retval)
{
	struct thread *thread;

	/* Try to find the thread */
	thread_atomic_start();
	thread = thread_sched_find(id);
	if (thread == NULL) {
		thread_atomic_stop();
		return (-1);
	}

	/* check its status */
	if (thread->scheduler.status != THREAD_STATUS_ZOMBIE) {
		thread_atomic_stop();
		return (-2);
	}

	/* destroy the thread */
	if (retval != NULL)
		*retval = (void*)((uintptr_t)thread->private.ret);
	thread_sched_remove(thread);
	free(thread);
	thread_atomic_stop();
	return (0);
}

/* thread_join(): Wait a thread terminason */
int thread_join(thread_t id, void **retval)
{
	while (thread_tryjoin(id, retval) != 0) {
		thread_kernel_yield();
	}
	return (0);
}

/* thread_exit(): Terminate the calling thread  */
void thread_exit(void *retval)
{
	struct thread *thread;

	/* get the current thread */
	thread_atomic_start();
	thread = thread_sched_get_current();
	if (thread == NULL) {
		thread_atomic_stop();
		while (1) { thread_kernel_yield(); };
	}

	/* Check thread longjmp() attribute feature */
	if (thread->attr.private.watermark == THREAD_ATTR_WATERMARK
	&&  thread->attr.state.longjmp == THREAD_ATTR_LONGJMP_ENABLE) {
		longjmp(thread->private.jmpbuf, 1);
	}

	/* block all signals */
	thread->signals.blocking = -1;
	thread->private.ret = (uintptr_t)retval;
	thread_signals_raise(thread, SIGTERM);
	thread_atomic_stop();

	/* Wait and yield */
	while (1) {
		thread_kernel_yield();
	}
}

/* thread_kill(): Send signals */
/* TODO: hierarchical handling ! */
int thread_kill(thread_t id, int sig)
{
	struct thread *thread;
	int ret;

	thread_atomic_start();
	thread = thread_sched_find(id);
	if (thread == NULL) {
		thread_atomic_stop();
		return (-1);
	}
	ret = thread_signals_raise(thread, sig);
	thread_atomic_stop();
	return (ret);
}

/* thread_signal(): set the disposition of the signal signum to handler */
void (*thread_signal(int signum, void (*handler)(int)))(int)
{
	struct thread *thread;
	void (*ret)(int);

	thread_atomic_start();
	thread = thread_sched_get_current();
	if (thread == NULL) {
		thread_atomic_stop();
		return (SIG_ERR);
	}
	ret = thread_signals_replace(thread, signum, handler);
	thread_atomic_stop();
	return (ret);
}

//---
//	Kernel interface
//---

/* thread_terminate(): Terminate a thread */
int thread_terminate(struct thread *thread)
{
	/* freed the stack */
	if (thread->private.stack != 0x00000000)
		free((void*)thread->private.stack);

	/* check is the thread is detached*/
	if (thread->attr.private.watermark == THREAD_ATTR_WATERMARK
	&&  thread->attr.state.detach == THREAD_ATTR_DETACHED) {
		thread_sched_remove(thread);
		free(thread);
		return (0);
	}

	/* the thread is joinable, wait until someone read its ret value
	   (the thread will be destroyed in the "thread_tryjoin()" function ) */
	thread->scheduler.status = THREAD_STATUS_ZOMBIE;
	return (1);
}

//---
//	Driver part
//---

/* init(): setup the scheduler */
static void init(void)
{
	/* initialize the scheduler */
	thread_sched_init();

	/* intialize the idle thread */
	thread_idle_init();
}

/* The thread scheduler is consider like a driver */
gint_driver_t drv_thread = {
	.name = "THREAD",
	.init = init
};
GINT_DECLARE_DRIVER(4, drv_thread);
